# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from cil.framework import Processor, DataContainer, AcquisitionData,\
 AcquisitionGeometry, ImageGeometry, ImageData
import numpy

class Normaliser(Processor):
    '''Normalization based on flat and dark
    
    This processor read in a AcquisitionData and normalises it based on 
    the instrument reading with and without incident photons or neutrons.
    
    Input: AcquisitionData
    Parameter: 2D projection with flat field (or stack)
               2D projection with dark field (or stack)
    Output: AcquisitionDataSetn
    '''
    
    def __init__(self, flat_field = None, dark_field = None, tolerance = 1e-5):
        kwargs = {
                  'flat_field'  : flat_field, 
                  'dark_field'  : dark_field,
                  # very small number. Used when there is a division by zero
                  'tolerance'   : tolerance
                  }
        
        #DataProcessor.__init__(self, **kwargs)
        super(Normaliser, self).__init__(**kwargs)
        if not flat_field is None:
            self.set_flat_field(flat_field)
        if not dark_field is None:
            self.set_dark_field(dark_field)
    
    def check_input(self, dataset):
        if dataset.number_of_dimensions == 3 or\
           dataset.number_of_dimensions == 2:
               return True
        else:
            raise ValueError("Expected input dimensions is 2 or 3, got {0}"\
                             .format(dataset.number_of_dimensions))
    
    def set_dark_field(self, df):
        if type(df) is numpy.ndarray:
            if len(numpy.shape(df)) == 3:
                raise ValueError('Dark Field should be 2D')
            elif len(numpy.shape(df)) == 2:
                self.dark_field = df
        elif issubclass(type(df), DataContainer):
            self.dark_field = self.set_dark_field(df.as_array())
    
    def set_flat_field(self, df):
        if type(df) is numpy.ndarray:
            if len(numpy.shape(df)) == 3:
                raise ValueError('Flat Field should be 2D')
            elif len(numpy.shape(df)) == 2:
                self.flat_field = df
        elif issubclass(type(df), DataContainer):
            self.flat_field = self.set_flat_field(df.as_array())
    
    @staticmethod
    def normalize_projection(projection, flat, dark, tolerance):
        a = (projection - dark)
        b = (flat-dark)
        with numpy.errstate(divide='ignore', invalid='ignore'):
            c = numpy.true_divide( a, b )
            c[ ~ numpy.isfinite( c )] = tolerance # set to not zero if 0/0 
        return c
    
    @staticmethod
    def estimate_normalised_error(projection, flat, dark, delta_flat, delta_dark):
        '''returns the estimated relative error of the normalised projection
        
        n = (projection - dark) / (flat - dark)
        Dn/n = (flat-dark + projection-dark)/((flat-dark)*(projection-dark))*(Df/f + Dd/d)
        ''' 
        a = (projection - dark)
        b = (flat-dark)
        df = delta_flat / flat
        dd = delta_dark / dark
        rel_norm_error = (b + a) / (b * a) * (df + dd)
        return rel_norm_error
        
    def process(self, out=None):
        
        projections = self.get_input()
        dark = self.dark_field
        flat = self.flat_field
        
        if projections.number_of_dimensions == 3:
            if not (projections.shape[1:] == dark.shape and \
               projections.shape[1:] == flat.shape):
                raise ValueError('Flats/Dark and projections size do not match.')
                
                   
            a = numpy.asarray(
                    [ Normalizer.normalize_projection(
                            projection, flat, dark, self.tolerance) \
                     for projection in projections.as_array() ]
                    )
        elif projections.number_of_dimensions == 2:
            a = Normalizer.normalize_projection(projections.as_array(), 
                                                flat, dark, self.tolerance)
        y = type(projections)( a , True, 
                    dimension_labels=projections.dimension_labels,
                    geometry=projections.geometry)
        return y
    


# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from cil.io import TIFFWriter
from cil.io import TIFFStackReader

from cil.framework import ImageGeometry, AcquisitionGeometry
import os
import numpy as np
import unittest
import shutil, glob

class TIFFReadWriter(unittest.TestCase):
    def setUp(self):
        ig = ImageGeometry(10,11,12, channels=3)

        data = ig.allocate(0)
        arr = data.as_array()
        k = 0
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                arr[i][j] += k
                k += 1
        data.fill(arr)
        self.cwd = os.getcwd()
        fname = os.path.join(self.cwd, 'test_tiff','myfile.tif')
        self.data_dir = os.path.dirname(fname)
        writer = TIFFWriter(data=data, file_name=fname, counter_offset=0)
        writer.write()
        
        self.ig = ig
        self.data = data
        
    def tearDown(self):
        shutil.rmtree(self.data_dir)

    def test_write_expected_num_files(self):
        data = self.data
        print (data.shape)
        print ("expecting {} files".format(data.shape[0]*data.shape[1]))
        files = glob.glob(os.path.join(self.data_dir, '*'))
        assert len(files) == data.shape[0]*data.shape[1]
    
    def test_read1(self):
        data = self.data
        reader = TIFFStackReader(file_name = self.data_dir)
        read = reader.read()
        np.testing.assert_array_equal(read.flatten(), data.as_array().flatten())
    
    def test_read_as_ImageData1(self):
        reader = TIFFStackReader(file_name = self.data_dir)
        
        img = reader.read_as_ImageData(self.ig)
        np.testing.assert_array_equal(img.as_array(), self.data.as_array())
    
    def test_read_as_ImageData_Exceptions(self):
        igs = [ ImageGeometry(10,11,12, channels=5) ]
        igs.append( ImageGeometry(12,32) )
        reader = TIFFStackReader(file_name = self.data_dir)
        
        for geom in igs:
            try:
                img = reader.read_as_ImageData(geom)
                assert False
            except ValueError as ve:
                print (ve)
                assert True

    def test_read_as_AcquisitionData1(self):
        ag = AcquisitionGeometry.create_Parallel3D()
        ag.set_panel([10,11])
        ag.set_angles([i for i in range(12)])
        ag.set_channels(3)
        print (ag.shape)
        # print (glob.glob(os.path.join(self.data_dir, '*')))
        reader = TIFFStackReader(file_name = self.data_dir)
        acq = reader.read_as_AcquisitionData(ag)

        np.testing.assert_array_equal(acq.as_array(), self.data.as_array())

    def test_read_as_AcquisitionData2(self):
        # with this data will be scrambled but reshape is possible
        ag = AcquisitionGeometry.create_Parallel3D()
        ag.set_panel([11,10])
        ag.set_angles([i for i in range(12)])
        ag.set_channels(3)

        reader = TIFFStackReader(file_name = self.data_dir)
        acq = reader.read_as_AcquisitionData(ag)

        np.testing.assert_array_equal(acq.as_array().flatten(), self.data.as_array().flatten())
    def test_read_as_AcquisitionData_Exceptions1(self):

            ag = AcquisitionGeometry.create_Parallel3D()
            ag.set_panel([11,12])
            ag.set_angles([i for i in range(12)])
            ag.set_channels(3)
            reader = TIFFStackReader(file_name = self.data_dir)
            try:
                acq = reader.read_as_AcquisitionData(ag)
                assert False
            except ValueError as ve:
                print (ve)
                assert True
    def test_read_as_AcquisitionData_Exceptions2(self):

            ag = AcquisitionGeometry.create_Parallel3D()
            ag.set_panel([11,12])
            ag.set_angles([i for i in range(12)])
            ag.set_channels(3)
            reader = TIFFStackReader(file_name = self.data_dir)

            try:
                class Fake(object):
                    pass
                fake = Fake()
                fake.shape = (36,11,10)
                acq = reader.read_as_ImageData(fake)
                assert False
            except TypeError as te:
                print (te)
                assert True

if __name__ == '__main__':
    ig = ImageGeometry(10,11,12, channels=3)

    data = ig.allocate(0)
    arr = data.as_array()
    k = 0
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            arr[i][j] += k
            k += 1
    data.fill(arr)
    os.mkdir("test_tiff")
    cwd = os.getcwd()
    fname = os.path.join(cwd, 'test_tiff','myfile.tif')
    data_dir = os.path.dirname(fname)
    writer = TIFFWriter(data=data, file_name=fname, counter_offset=0)
    writer.write()

    ag = AcquisitionGeometry.create_Parallel3D()
    ag.set_panel([10,11])
    ag.set_angles([i for i in range(12)])
    ag.set_channels(3)
    print (ag.shape)
    print (glob.glob(os.path.join(data_dir, '*')))
    reader = TIFFStackReader(file_name = os.path.dirname(fname))
    acq = reader.read_as_AcquisitionData(ag)

    print (type(acq), type(data))
    np.testing.assert_array_equal(acq.as_array(), data.as_array())


# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from cil.utilities.dataexample import TestData

def gaussian(image, seed=None, clip=True, **kwargs):
    '''Gaussian-distributed additive noise.
    
        seed : int, optional
            If provided, this will set the random seed before generating noise,
            for valid pseudo-random comparisons.
        clip : bool, optional
            If True (default), the output will be clipped after noise applied. 
            This is needed to maintain the proper image data range. If False, clipping
            is not applied, and the output may extend beyond the range [-1, 1].
        mean : float, optional
            Mean of random distribution.
            Default : 0.
        var : float, optional
            Variance of random distribution. 
            Note: variance = (standard deviation) ** 2. Default : 0.01
        
        '''
    return TestData.random_noise(image, mode='gaussian', seed=seed, clip=clip, **kwargs)

def poisson(image, seed=None, clip=True, **kwargs):
    '''Poisson-distributed noise generated from the data.
    
        seed : int, optional
            If provided, this will set the random seed before generating noise,
            for valid pseudo-random comparisons.
        clip : bool, optional
            If True (default), the output will be clipped after noise applied. 
            This is needed to maintain the proper image data range. If False, clipping
            is not applied, and the output may extend beyond the range [-1, 1].
        
        '''
    return TestData.random_noise(image, mode='poisson', seed=seed, clip=clip, **kwargs)

def salt(image, seed=None, **kwargs):
    '''Replaces random pixels with 1.
    
        seed : int, optional
            If provided, this will set the random seed before generating noise,
            for valid pseudo-random comparisons.
        amount : float, optional
            Proportion of image pixels to replace with noise on range [0, 1].
            Used in 'salt', 'pepper', and 'salt & pepper'. Default : 0.05
        '''
    return TestData.random_noise(image, mode='salt', seed=seed, clip=True, **kwargs)

def pepper(image, seed=None, **kwargs):
    '''Replaces random pixels with 0 (for unsigned images) or -1 (for signed images).
    
        seed : int, optional
            If provided, this will set the random seed before generating noise,
            for valid pseudo-random comparisons.
        amount : float, optional
            Proportion of image pixels to replace with noise on range [0, 1].
            Used in 'salt', 'pepper', and 'salt & pepper'. Default : 0.05
        '''
    return TestData.random_noise(image, mode='pepper', seed=seed, clip=True, **kwargs)

def saltnpepper(image, seed=None, **kwargs):
    '''Replaces random pixels with either 1 or `low_val`
    
    `low_val` is 0 for unsigned images or -1 for signed images.
    
        seed : int, optional
            If provided, this will set the random seed before generating noise,
            for valid pseudo-random comparisons.
        amount : float, optional
            Proportion of image pixels to replace with noise on range [0, 1].
            Used in 'salt', 'pepper', and 'salt & pepper'. Default : 0.05
        salt_vs_pepper : float, optional
            Proportion of salt vs. pepper noise for 's&p' on range [0, 1].
            Higher values represent more salt. Default : 0.5 (equal amounts)'''
    return TestData.random_noise(image, mode='s&p', seed=seed, clip=True, **kwargs)

def speckle(image, seed=None, clip=True, **kwargs):
    '''Multiplicative noise
    
    using out = image + n*image, where
                        n is uniform noise with specified mean & variance.
                        
    seed : int, optional
            If provided, this will set the random seed before generating noise,
            for valid pseudo-random comparisons.
        clip : bool, optional
            If True (default), the output will be clipped after noise applied.
            This is needed to maintain the proper image data range. If False, clipping
            is not applied, and the output may extend beyond the range [-1, 1].
        mean : float, optional
            Mean of random distribution.
            Default : 0.
        var : float, optional
            Variance of random distribution.
            Note: variance = (standard deviation) ** 2. Default : 0.01
        '''
    return TestData.random_noise(image, mode='speckle', seed=seed, clip=clip, **kwargs)

def localvar(image, seed=None, clip=True, **kwargs):
    '''Gaussian-distributed additive noise, with specified
                        local variance at each point of `image`.
                        
        seed : int, optional
            If provided, this will set the random seed before generating noise,
            for valid pseudo-random comparisons.
        clip : bool, optional
            If True (default), the output will be clipped after noise applied. This is
            needed to maintain the proper image data range. If False, clipping
            is not applied, and the output may extend beyond the range [-1, 1].
        mean : float, optional
            Mean of random distribution. Used in 'gaussian' and 'speckle'.
            Default : 0.
        var : float, optional
            Note: variance = (standard deviation) ** 2. Default : 0.01
        local_vars : ndarray, optional
            Array of positive floats, same shape as `image`, defining the local
            variance at every image point. Used in 'localvar'.
        '''
    return TestData.random_noise(image, mode='localvar', seed=seed, clip=clip, **kwargs)